/**
 * The scenario:
 * A video server is on a wired network. It is connected through a 1Gbps links to a set of RSU. 
 * A set of OBU are moving along the x-axis with a mobility pattern describe in mobility trace file.
 * The video server opens connection to each OBU and send videos (one per OBU) according to a 
 * video trace file. 
 * At the end of the simulation, we print the number of downloaded MB and the disruption time.
 * 
 * If you use this ns-3 file, plkease cite the following paper:
 * Thomas Begin, Anthony Busson, Isabelle Guerin-Lassous, Azzedine Boukerche
 * Video on Demand in IEEE 802.11p-based Vehicular Networks: Analysis and Dimensioning
 * MSWIM 2018. 
 */

/*** Most of the comments are in french...sorry	*/


#include "ns3/vector.h"
#include "ns3/string.h"
#include "ns3/socket.h"
#include "ns3/double.h"
#include "ns3/config.h"
#include "ns3/log.h"
#include "ns3/command-line.h"
#include "ns3/mobility-model.h"
#include "ns3/yans-wifi-helper.h"
#include "ns3/position-allocator.h"
#include "ns3/mobility-helper.h"
#include "ns3/mobility-module.h"


#include "ns3/internet-stack-helper.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/ipv4-interface-container.h"
#include <iostream>

#include "ns3/ocb-wifi-mac.h"
#include "ns3/wifi-80211p-helper.h"
#include "ns3/wave-mac-helper.h"
#include "ns3/wifi-net-device.h"

//CSMA
#include <iostream>
#include <fstream>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/csma-module.h"
#include "ns3/applications-module.h"
#include "ns3/internet-module.h"


//OLSR
#include "ns3/olsr-helper.h"
#include "ns3/olsr-routing-protocol.h"
#include "ns3/ipv4-list-routing-helper.h"

//To parse the video trace file 
#include <fstream> 

#define DEBUG false
#define PCAP false

#define MSS 1440 //Au de-la de 1440 il fragmente...

using namespace ns3;

uint16_t dstport = 12345;
float timeBeforeStartingVideo=0.0;//Time between the reception of the first frame and the time the video is played
float distanceRSU=2000.0;	  //Distance between RSU
float tempsFinSimu=60.0;	  

NS_LOG_COMPONENT_DEFINE ("WifiSimpleOcb");


/********************************************************************************************************/
/*													*/
/*			Class videoFile: open and parse the video trace file.				*/
/*							 						*/
/*													*/
/*													*/
/********************************************************************************************************/

//Constant used to parse the video trace file
const int MAX_CHARS_PER_LINE = 512;
const int MAX_TOKENS_PER_LINE = 20;
const char* const DELIMITER = "\t";//To modify as function of the trace file format


class videoFile{

  std::string fileName;//Nom du fichier 
  std::ifstream video;//fichier 
 
  int lineNumber=0;		//Compte le nombre de lignes lues (pour le Debug)
  int frameNumber=0;		//Numero de seq de la trame en emission (c'est le numero dans le fichier de trace video)
  uint32_t frameSize=0;		//La taille totale de la trame (celle du fichier)
  double frameTime=0.0;		//Temps de reconstitution de la trame lors de la visualisation de la video

  public:
    videoFile();		//Constructeur: Ouvre le fichier starWars	
    videoFile(std::string);	//Constructeur: Ouvre le fichier indique
    int openVideo();		//Ouvre la video "fileName"
    int openVideo(std::string);	//Ouvre la video en specifiant un nom (change le nom fileName dans la classe)
    int getNextFrame();		//Lit la prochaine ligne du fichier
    uint32_t getFrameSize();	//Renvoi frameSize
    double getFrameTime();	//Renvoi frameTime
    std::string getName();	//Renvoi fileName
};
  

//Constructeurs - Node et nom de la trace video + ouverture
videoFile::videoFile()
{
  fileName="videoTrace/starWars.txt";
}

videoFile::videoFile (std::string name)
{
  fileName=name;
}

//Ouvre la video
int videoFile::openVideo()
{
  video.open(fileName); // open the file
  if (!video.good())
  {
    NS_LOG_UNCOND ("Error: Video trace not found - abort or error...");
    exit(1);
  }
  if(DEBUG) std::cout << "videoStream: file " << fileName << " open!" << std::endl;
  return(0);
}

//Ouvre la video en specifiant un nom
int videoFile::openVideo(std::string name)
{
  fileName=name;
  return(openVideo()); // open the file
}

//Read a line in the video trace file. 
//On boucle sur la partie commentaire jusqu'a arrivee sur la premiere trame video. 
//Return 0 si en fin de fichier
//return 1 si on est pas en fin de fichier
//Les erreurs ne sont pas gerees. 
int videoFile::getNextFrame()
{
  bool comment=true;
  char buf[MAX_CHARS_PER_LINE];

  while(comment)
  {
    if (video.getline(buf, MAX_CHARS_PER_LINE))
    {
      lineNumber++;
      // parse the line into blank-delimited tokens
      int n = 0; // a for-loop index
    
      // array to store memory addresses of the tokens in buf
      const char* token[MAX_TOKENS_PER_LINE] = {}; // initialize to 0
    
      // parse the line
      token[0] = strtok(buf, DELIMITER); // first token
      if (token[0]) // zero if line is blank
      {
        for (n = 1; n < MAX_TOKENS_PER_LINE; n++)
        {
          token[n] = strtok(0, DELIMITER); // subsequent tokens
          if (!token[n]) break; // no more tokens
        }
      }

      //On compare le 1er caractere a "#" -> version C
      if( *token[0] == '#')
      {
        if(DEBUG) std::cout << "C'est un commentaire!!!!" << std::endl;
      } else {
	comment=false;
      }

      if(!comment)
      {
        //A modifier en fonction du format de fichier
        frameNumber=atoi(token[0]);
        frameTime=atof(token[1]);
	frameSize=atoi(token[3]);
        if(DEBUG) std::cout << "DEBUG getNextVideoFrame:" << frameNumber << " Size: " << frameSize << " Time: " << frameTime << std::endl;//DEBUG
      }	
      // process (print) the tokens - DEBUG
      /*
      for (int i = 0; i < n; i++) // n = #of tokens
        std::cout << "Token[" << i << "] = " << token[i] << std::endl;
      std::cout << std::endl;
      */

    }//fin du if !file.eof()
    else
    {
      if(DEBUG) std::cout << "DEBUG getNextVideoFrame: fin de fichier (" << lineNumber << " lignes)\n";
      return(0);//return 0: end of file. 
    }
  }

  return(1);
}

uint32_t videoFile::getFrameSize()
{
	return(frameSize);
}

double videoFile::getFrameTime()
{
	return(frameTime);
}

std::string videoFile::getName()
{
	return(fileName);
}

/********************************************************************************************************/
/*													*/
/*				Class videoPlayer a installer sur les OBUs 				*/
/*			Gere toute la partie socket et la partie lecture des trames.			*/
/*													*/
/*													*/
/********************************************************************************************************/

class videoPlayer
{
  videoFile	video;
  Ptr< Node >  	node;		//Noeud sur lequel on installe le player
  Ipv4Address   dstaddr;	//Adresse IP destination
  Ptr<Socket>   socketPlayer;	//Socket sur laquelle on lit la video
  bool 		videoDone=false;

  //Gestion de la lecture
  double 	videoTime=0.0;		//Temps de lecture de la video
  double 	startTime=-1.0;		//Temps du debut de la lecture: x sec apres la reception du premier paquet
  double	timeWithoutVideo=0.0; 	//Temps durant lequel on a pas de video
  uint32_t	bytesRecv=0;  	//Nb d'octets recu pour la trame en cours 		
  uint32_t 	totalRecv=0; 	//Volume total recu (en octets)
  uint32_t	lastCall=0; //Var entiere qui sert a afficher un log qu'une seule fois. Meriterai d'etre declare ailleurs...
  uint32_t	done=0;     //idem

  public:
    videoPlayer(Ptr<Node>);				//Constructeur
    videoPlayer(Ptr<Node>, std::string);		//Constructeur
    void openSocket();
    void receivePacket(Ptr<Socket>);
    void accept(Ptr<Socket>, const ns3::Address&);
    void checkVideoBuffer();
    bool updateVideoPlayer(uint32_t);
};

//Constructeur
videoPlayer::videoPlayer(Ptr<Node> theNode)
{
  node=theNode;
  video.openVideo();
  openSocket();
}

videoPlayer::videoPlayer(Ptr<Node> theNode, std::string fileName)
{
  node=theNode;
  video.openVideo(fileName);
  openSocket();
}

void videoPlayer::openSocket()
{
  //Creation de la socket TCP (le lecteur / OBU)
  TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
  socketPlayer = Socket::CreateSocket (node, tid);//Destination-recepteur


  InetSocketAddress local = InetSocketAddress (Ipv4Address::GetAny (), dstport);//Attention au port si mis a jour par ailleurs
  socketPlayer->Bind (local);
  socketPlayer->Listen ();

	
  socketPlayer->SetAcceptCallback ( MakeNullCallback<bool, Ptr< Socket >, const Address &> (), MakeCallback (&videoPlayer::accept, this));

}

void videoPlayer::accept(Ptr<Socket> socket, const ns3::Address& from)
{

  if(DEBUG) std::cout << "Connection accepted" << std::endl;
  video.getNextFrame();		//On charge la toute premiere trame
  socket->SetRecvCallback (MakeCallback (&videoPlayer::receivePacket,this));

}

//checkVideoBuffer calcul le temps durant lequel il n'y a pas de video a lire (la video est en pause en attente de données).
//Appel recursif
void videoPlayer::checkVideoBuffer()
{
  double bufferTime=video.getFrameTime()+startTime;//On a des images jusqu'a cette date 
  double nextCall; 


  if(DEBUG) std::cout << "OBU " << node->GetId() << ": checkVideoBuffer. bufferTime= " << bufferTime << " startTime= " << startTime << " Now=" << Simulator::Now ().GetSeconds () << " timeWithoutVideo=" << timeWithoutVideo << std::endl;

  //Log utilise pour mon calcul
  if(lastCall==1 && done==0)
  { 
    std::cerr << distanceRSU << " " << node->GetId() << " " << Simulator::Now ().GetSeconds () << " " << totalRecv/1000000.0 << " " << bufferTime << " " << timeWithoutVideo << std::endl;
    done=1; 
  }

  if(Simulator::Now ().GetSeconds () >= bufferTime)//Pas de video
  {
    timeWithoutVideo+=Simulator::Now ().GetSeconds () - bufferTime;
    startTime+=Simulator::Now ().GetSeconds () - bufferTime;//Pour eviter d'avoir a rattraper le retard dans le calcul (ToDo: Verifier).  
    nextCall=0.5; //On rappelle cette fonction dans 0.5 sec
  }else{
    nextCall=bufferTime-Simulator::Now ().GetSeconds (); //On rappelle cette fonction au moment ou le tampon aura ete vide	
    if(nextCall<0.1) nextCall=0.1; //Petite securite pour eviter de reboucler indefiniment.
  }
  //On appelle au plus tard a l'echeance de la simulation
  if( done==0 && lastCall==0 && (nextCall >= (tempsFinSimu-Simulator::Now ().GetSeconds())) ) 
  {
    nextCall=tempsFinSimu-Simulator::Now ().GetSeconds ();
    if(nextCall<0.0) nextCall=0.0;
    lastCall=1;
  } else lastCall=0;

  nextCall+=(node->GetId()/100.0);//Pour desynchroniser

  if(!videoDone) Simulator::Schedule (Seconds(nextCall), &videoPlayer::checkVideoBuffer, this);
  else if(DEBUG) std::cout << "OBU " << node->GetId() << ": the video has been stopped for " << timeWithoutVideo << " sec.\n";

}

//L'argument "socket" ne sert a rien en pratique ici mais les formats des callbacks sont fixes et supposent une socket en argument. 
void videoPlayer::receivePacket (Ptr<Socket> socket)
{
  //Appel non bloquant. Il va boucler jusqu'a qu'il n'y ait plus de données a lire.  
  //Il faut un callback (voir plus haut) pour appeler cette fonction que lorsque des donnees sont disponibles. 
  Ptr<Packet> packet;
  bool videoEnd=false;


  while ( !videoEnd && (packet=socket->Recv ())) 
  {
    if(DEBUG) std::cout << "Recu 1 paquet de " << packet->GetSize() << std::endl;
    videoEnd=updateVideoPlayer(packet->GetSize());
  }

  //Tout premier paquet: on met a jour le temps a partir duquel la video demarre (pour le passager). 
  if(startTime<0.0)
  {	
	startTime=Simulator::Now ().GetSeconds ()+timeBeforeStartingVideo; //Voir si on modifie 5.0 
	checkVideoBuffer();
  }

  if(videoEnd)
  {
    if(DEBUG) std::cout << "OBU " << node->GetId() << ": the video has been received. We close the socket.\n";
    videoDone=true;
  }
}



bool videoPlayer::updateVideoPlayer(uint32_t packetSize)
{
	bool videoEnd=false; 

	bytesRecv+=packetSize;
	totalRecv+=packetSize;

	if(DEBUG) std::cout << "OBU " << node->GetId() << " VideoPlayer: " <<	bytesRecv << " received (packet=" << packetSize << ") over " <<  video.getFrameSize() << " expected (totalRecv= " << totalRecv << ") at " << Simulator::Now().GetSeconds() << std::endl;  

	//On teste si la trame a ete recu au complet (il peut y en avoir plusieurs dans le meme paquet d'ou le while)
	//videoTime est le temps de reconstitution relatif (temps trame video par rapport a la date de debut de lecture: 0)
	while( !videoEnd && bytesRecv >= video.getFrameSize())
	{
		videoTime=video.getFrameTime(); //On avance dans la lecture de la video
		bytesRecv-=video.getFrameSize();//On retire la taille de la trame de ce qui a ete recu
		if(DEBUG) std::cout << "VideoPlayer: a full frame has been loaded - videoTime=" << videoTime << std::endl;  
		//On charge les proprietes de la trame suivante 
		if(video.getNextFrame()==1) 
		{ 
			if(DEBUG) std::cout << "VideoPlayer: next frame has size=" << video.getFrameSize() << std::endl;  
		} else videoEnd=true;
	}
	return(videoEnd);
}


/********************************************************************************************************/
/*													*/
/*				Class du Video Server videoServerApp					*/
/*													*/
/*													*/
/*													*/
/********************************************************************************************************/


class videoServerApp{

  videoFile	video;
  Ptr< Node >  	node;//Noeud sur lequel on installe le server
  Ipv4Address   dstaddr;//Adresse IP destination
  
  //Gestion du flux
  uint32_t currentTxBytes = 0;	//Ce qui a deja ete emis (attention sur la socket et pas forcement sur le reseau)
  uint32_t bytesToSend = 0;	//Taille total de la video

  public: 
    videoServerApp(Ptr< Node > );//Constructeur sans le nom / starwars par defaut
    videoServerApp(Ptr< Node >, std::string);//Constructeur avec le nom du fichier
    videoServerApp(Ptr< Node >, const char*);//Constructeur avec l'adresse IP de l'Obu
    videoServerApp(Ptr< Node >, std::string, const char*);//Constructeur avec le nom du fichier et l'adresse IP destination

    int openVideo();
    void testConnectivity (Ptr<Socket> localSocket);
    void WriteUntilBufferFull (Ptr<Socket> localSocket, uint32_t txSpace);
    void successConnect(Ptr<Socket> localSocket);
    void connectAttempt(Ptr<Socket> localSocket);
    void connectFails(Ptr<Socket> localSocket);
    void reconnect(Ptr<Socket> localSocket);
    void connectToObu();
    uint32_t getVideoSize();


};//fin de la class videoServer


//Constructeurs - Node et nom de la trace video et adresse IP obu + ouverture
videoServerApp::videoServerApp(Ptr< Node > theNode)
{
  node=theNode;
  video.openVideo();
  bytesToSend=getVideoSize();
  dstaddr.Set("10.2.0.3");
}

videoServerApp::videoServerApp(Ptr< Node > theNode, std::string name)
{
  node=theNode;
  video.openVideo(name);
  bytesToSend=getVideoSize();
  dstaddr.Set("10.2.0.3");
}

videoServerApp::videoServerApp(Ptr< Node > theNode, const char* ipDest)
{
  node=theNode;
  video.openVideo();
  bytesToSend=getVideoSize();
  dstaddr.Set(ipDest);
}

videoServerApp::videoServerApp(Ptr< Node > theNode, std::string name, const char* ipDest)
{
  node=theNode;
  video.openVideo(name);
  bytesToSend=getVideoSize();
  dstaddr.Set(ipDest);
}

//Calcul la taille de la video
//Parcours le fichier et le ferme!!!
uint32_t videoServerApp::getVideoSize()
{
  uint32_t size=0;
  while(video.getNextFrame()==1)
  {
	size+=video.getFrameSize();
  }

  if(DEBUG) std::cout << "DEBUG: video size " << size << std::endl;
  return(size);
}

//Lance des paquets tous les x secondes
//Attention dans le scenario mobile, le noeud emetteur a bien avance (en terme de N° seq TCP) quand le vehicule est entre les deux OBUs. 
//Quand il arrive sur le 2nd OBU, il attand 7 ACK DUP avant de renvoyer le manquant (voir wireshark). Si le temps entre deux tests est de 3sec cela mettra 21 sec avant
//de renvoyer les segments manquants. 
void videoServerApp::testConnectivity (Ptr<Socket> localSocket)
{
  
  uint8_t data[1000];
  double waitingBeforeTrying=3.0;


  if(DEBUG) std::cout << "Simulator time: " << Simulator::Now ().GetSeconds () << std::endl;
  if (currentTxBytes < 100)
    {
      int amountSent = localSocket->Send (&data[0], 10, 0);

      if(DEBUG) std::cout << "Test connectivity: send 10 bytes / amountSent = " << amountSent << std::endl;
      currentTxBytes += amountSent;

      Simulator::Schedule (Seconds(waitingBeforeTrying), &videoServerApp::testConnectivity, this, localSocket);
    } else localSocket->Close(); 
}

void videoServerApp::WriteUntilBufferFull (Ptr<Socket> localSocket, uint32_t txSpace)//On doit laisser deux arguments a cause du callBack
{
  
  //MSS de xx octets max declare en const gobal en haut de ce fichier
  uint8_t data[MSS];

  while (currentTxBytes < bytesToSend && localSocket->GetTxAvailable () > 0)
  {
	
      uint32_t toWrite = std::min (MSS, (int) localSocket->GetTxAvailable ());//A surveiller
      toWrite = std::min (toWrite, bytesToSend-currentTxBytes);//Minimum du MSS, de la taille du buffer d'emission et du nombre d'octets restant a transmettre. 

      int amountSent = localSocket->Send (&data[0], toWrite, 0);
      if(DEBUG) std::cout << "WriteUntil: send " << toWrite << " bytes / currentTxBytes = " << currentTxBytes << " totalVideoSize=" << bytesToSend <<  std::endl;

      if(amountSent < 0)
      {
        // we will be called again when new tx space becomes available.
	if(DEBUG) std::cout << "Erreur de transmission Send fails!\n";
      } else {
      	currentTxBytes += amountSent;
      }
      if(DEBUG) std::cout << "Amount sent: " << amountSent << std::endl;
  }

  if(currentTxBytes >= bytesToSend)
  {
    if(DEBUG) std::cout << "Video server: the full video has been sent. We close the socket." << std::endl; 
    localSocket->SetSendCallback ( MakeNullCallback<void, Ptr<Socket>, uint32_t> () );//Annule le callBack
    localSocket->Close ();
  }
}

void videoServerApp::successConnect(Ptr<Socket> localSocket)
{
  // tell the tcp implementation to call WriteUntilBufferFull again
  // if we blocked and new tx buffer space becomes available

  if(DEBUG) std::cout << "Connect successfull\n";
  if(DEBUG) std::cout << "Begin to send " << bytesToSend << " bytes for the video " << video.getName() << "\n";


  //Le callback ci-dessous sert a se reconnecter a l'OBU en cas d'erreur de connexion TCP (coupure typiquement ici a cause du transit entre 2 OBU).

  //Call back sur l'emission
  Callback<void, Ptr<Socket>, uint32_t> callbackBufferFull = MakeCallback (&videoServerApp::WriteUntilBufferFull, this);

  //On envoit le plus qu'on peut
  localSocket->SetSendCallback (callbackBufferFull);//Il y a aussi un callBack sur l'emission setDataSendallBack --> VOIR
  WriteUntilBufferFull (localSocket, localSocket->GetTxAvailable ());

}

void videoServerApp::connectToObu()
{
  TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
  Ptr<Socket> localSocket = Socket::CreateSocket (node, tid);//Source

  connectAttempt(localSocket);

}

void videoServerApp::connectAttempt(Ptr<Socket> localSocket)
{

  Callback<void, Ptr<Socket> > callbackSuccess = MakeCallback (&videoServerApp::successConnect, this);


  localSocket->SetConnectCallback (callbackSuccess, MakeNullCallback<void, Ptr<Socket>> ()) ; 

  Callback<void, Ptr<Socket> > callbackClose = MakeCallback (&videoServerApp::reconnect, this);
  localSocket->SetCloseCallbacks ( MakeNullCallback<void, Ptr<Socket>> (), callbackClose); 	


  localSocket->Connect (InetSocketAddress (dstaddr, dstport)); //connect

  if(DEBUG)
  {
    if(localSocket->GetErrno()==Socket::ERROR_NOTERROR) 
    { 
  	if(DEBUG) std::cout << "Pas d'erreur de connexion\n";
    }
  }

  if(localSocket->GetErrno()==Socket::ERROR_NOROUTETOHOST) //C est lui qui ne marche pas: pas mis a jour????
  {
	localSocket->Close();
        
  	double waitingBeforeTrying=1.0;
	if(DEBUG) std::cout << "Erreur de connection (No route to host): on reprogramme dans une seconde\n";
        Simulator::Schedule (Seconds(waitingBeforeTrying), &videoServerApp::connectToObu, this);//On rappelle cette fct en cas de connection rate (ce sera une nouvelle socket: important). 

  }

}

void videoServerApp::reconnect(Ptr<Socket> localSocket)
{

  TypeId tid = TypeId::LookupByName ("ns3::TcpSocketFactory");
  localSocket = Socket::CreateSocket (node, tid);//Source

  Callback<void, Ptr<Socket> > callbackSuccess = MakeCallback (&videoServerApp::successConnect, this);

  localSocket->SetConnectCallback (callbackSuccess, MakeNullCallback<void, Ptr<Socket>> ()) ; 

  Callback<void, Ptr<Socket> > callbackClose = MakeCallback (&videoServerApp::connectAttempt, this);
  localSocket->SetCloseCallbacks ( MakeNullCallback<void, Ptr<Socket>> (), callbackClose); 	

  localSocket->Connect (InetSocketAddress (dstaddr, dstport)); //connect

  if(DEBUG)
  {
    if(localSocket->GetErrno()==Socket::ERROR_NOTERROR) 
    { 
  	if(DEBUG) std::cout << "Pas d'erreur de connexion\n";
    }
  }

  if(localSocket->GetErrno()==Socket::ERROR_NOROUTETOHOST) //C est lui qui ne marche pas: pas mis a jour????
  {
	localSocket->Close();
        
  	double waitingBeforeTrying=1.0;
	if(DEBUG) std::cout << "Erreur de connection (No route to host): on reprogramme dans une seconde\n";
        Simulator::Schedule (Seconds(waitingBeforeTrying), &videoServerApp::connectToObu, this);//On rappelle cette fct en cas de connection rate (ce sera une nouvelle socket: important). 

  }

}

/********************************************************************************************************/
/*													*/
/*				Fin de la Class Video Server						*/
/*													*/
/********************************************************************************************************/


int main (int argc, char *argv[])
{
  std::string phyMode ("OfdmRate6MbpsBW10MHz"); //Not used
  bool verbose = false;

  int nbOfRsu=3; //Attention: le fichier decrivant la mobilite depend du nombre d'OBU (A MODIFIER EN CONSEQUENCE)
  std::string videoName="videoTrace/starWars.txt";

  CommandLine cmd;

  cmd.AddValue ("distanceRSU", "Distance between RSU", distanceRSU);
  cmd.AddValue ("phyMode", "Wifi Phy mode", phyMode);
  cmd.AddValue ("videoName", "Name of the video", videoName);
  cmd.AddValue ("verbose", "turn on all WifiNetDevice log components", verbose);

  cmd.Parse (argc, argv);

  //Create the server, rsu (fixed 802.11p node - router between the wired and wireless networks), and obu. 
  NodeContainer obu;
  NodeContainer rsu;
  NodeContainer videoServer;


  videoServer.Create (1);
  rsu.Create (nbOfRsu);
  obu.Create (5);

  //Premier vehicule sort du dernier OBU
  tempsFinSimu=(1200.0+(nbOfRsu-1)*distanceRSU+700.0)/30.0;


  /************   Couche PHY-MAC 802.11: OBU et RSU     ***********************************/

  YansWifiPhyHelper wifiPhy =  YansWifiPhyHelper::Default ();
  YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default ();
  Ptr<YansWifiChannel> channel = wifiChannel.Create ();
  wifiPhy.SetChannel (channel);

  //Transmission power: ce que l'on obtient avec les valeurs par defaut c'est 130-150m de portee radio (trop court). 
  //802.11p uses different transmit powers for the service channels (SCH) and the control channel (CCH) i.e., CCH at 44.8dBm and SCH can vary from 23dBm to 40dBm. 
  //Src: http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=5547184
  //Ici on est en SCH (non critical application) donc entre 23 et 40dBm. Ci-dessous tout est en dBm (le gain aussi).
  //23dBm --> 220m de portee radio environ
  //38.5dBm --> 700-720m de portee radio environ
  //40dBm --> 800m de portee radio environ
  wifiPhy.Set ("TxPowerStart", DoubleValue(38.5));
  wifiPhy.Set ("TxPowerEnd", DoubleValue(38.5));
  wifiPhy.Set ("TxPowerLevels", UintegerValue(1));//1 seul valeur possible - pas de controle de puissance
  wifiPhy.Set ("TxGain", DoubleValue(0) );
  wifiPhy.Set ("RxGain", DoubleValue (0) );


  // ns-3 supports generate a pcap trace
  wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11);
  NqosWaveMacHelper wifi80211pMac = NqosWaveMacHelper::Default ();
  Wifi80211pHelper wifi80211p = Wifi80211pHelper::Default ();
  if (verbose)
    {
      wifi80211p.EnableLogComponents ();      // Turn on all Wifi 802.11p logging
    }

  //Constant transmission rate
  //wifi80211p.SetRemoteStationManager ("ns3::ConstantRateWifiManager", "DataMode",StringValue (phyMode), "ControlMode",StringValue (phyMode));
  
  //Adaptive transmission rate
  	//rtsThreshold: semble necessaire pour affecter le manager WiFi - a declarer uniquement dans ce cas
  uint32_t rtsThreshold = 999999; //  a value of 999999 disables it even for large A-MPDU 
  wifi80211p.SetRemoteStationManager ("ns3::IdealWifiManager", "RtsCtsThreshold", UintegerValue (rtsThreshold));

  NodeContainer cont80211p = NodeContainer (rsu, obu); //Conteneur commun pour cette technologie (ne se voit pas autrement si affecter séparément).
  NetDeviceContainer devices80211p = wifi80211p.Install (wifiPhy, wifi80211pMac, cont80211p);


  // Tracing wireshark
  if(PCAP) wifiPhy.EnablePcap ("wave-simple-80211p", devices80211p);



  /************   Couche PHY-MAC CSMA: RSU et serveur video     ***********************************/
  CsmaHelper csma;
  csma.SetChannelAttribute ("DataRate", DataRateValue (1000000000));//1Gbps
  csma.SetChannelAttribute ("Delay", TimeValue (MicroSeconds (1.0)));//Delay 1 microsec

  NodeContainer csmaC = NodeContainer (videoServer.Get (0), rsu);
  NetDeviceContainer devicesCsma = csma.Install (csmaC);

  // Tracing wireshark
  if(PCAP) csma.EnablePcap ("csma-trace", devicesCsma);


  /***************************** Mobility and position *************************************/

  //RSU position
  MobilityHelper mobilityRSU;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();

  //Scenario: tous les x km (0, distanceRSU, 2xdistanceRSU, etc.) - distanceRSU est passe en ligne de commande

  for (int i=0; i < nbOfRsu; i++)
  {
    positionAlloc->Add (Vector (1.0*i*distanceRSU, 0.0, 0.0));
  }
  mobilityRSU.SetPositionAllocator (positionAlloc);
  mobilityRSU.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobilityRSU.Install (rsu);

  //OBU position (it moves). Trace mobilite Ns2 dans le dossier mobilityTrace
  //Les details du format de fichier et des informations essentielles se trouvent dans "mobilityTrace/traceFile.ns_movements"
  std::string traceFile = "mobilityTrace/scenario1.ns_movements";
  Ns2MobilityHelper ns2 = Ns2MobilityHelper (traceFile);
  ns2.Install (); // configure movements for each node, while reading trace file


  /*******************************  IP and Routing  ******************************************/

  //Installation de la pile IP et assignation des adresses
  InternetStackHelper internet;

  double helloInterval=0.5; 
  double tcInterval=0.5;
  double hnaInterval=0.5;

  OlsrHelper olsr;
  olsr.Set ("HelloInterval",TimeValue(Seconds(helloInterval)));
  olsr.Set ("TcInterval",TimeValue(Seconds(tcInterval)));
  olsr.Set ("HnaInterval",TimeValue(Seconds(hnaInterval)));

  Ipv4ListRoutingHelper list;
  Ipv4StaticRoutingHelper staticRouting;

  list.Add (staticRouting, 0); 
  list.Add (olsr, 10); 
  internet.SetRoutingHelper (list);


  internet.Install (rsu);
  internet.Install (videoServer);

  /* OLSR pour OBU ou le forwading est proscrit (willingness=0) */
  InternetStackHelper internet1;
  OlsrHelper olsr1;
  olsr1.Set ("HelloInterval",TimeValue(Seconds(helloInterval)));
  olsr1.Set ("TcInterval",TimeValue(Seconds(tcInterval)));
  olsr1.Set ("HnaInterval",TimeValue(Seconds(hnaInterval)));
  olsr1.Set ("Willingness",EnumValue(0));//0: OLSR_WILL_NEVER

  Ipv4ListRoutingHelper list1;
  Ipv4StaticRoutingHelper staticRouting1;

  list1.Add (staticRouting1, 0); 
  list1.Add (olsr1, 10); 	
  internet1.SetRoutingHelper (list1);

  internet1.Install (obu);


  //Les adresses 
  //Server 10.1.0.1 et RSU cote filaire 10.1.0.2, 10.1.0.3, etc. 
  //RSU on the wireless network: 10.2.0.0/24: 10.2.0.1, 10.2.0.2, etc.
  //puis OBU: 10.2.0.RSU+1: 10.2.0.RSU+2, etc. Si il y a 10 RSU, le premier OBU a pour adresse: 10.2.0.11

  //Server and RSU on the Csma network
  Ipv4AddressHelper ipv4Csma;
  NS_LOG_INFO ("Assign IP Addresses.");
  ipv4Csma.SetBase ("10.1.0.0", "255.255.0.0");
  Ipv4InterfaceContainer i = ipv4Csma.Assign (devicesCsma); //Server and RSU in the CSMA network

  //RSU et OBU sur le 802.11p  
  Ipv4AddressHelper ipv480211p;
  ipv480211p.SetBase ("10.2.0.0", "255.255.0.0");//65500 adresses possibles 
  Ipv4InterfaceContainer i1 = ipv480211p.Assign (devices80211p);


  //Valeur par defaut de la classe TCP: le change pour toute les sockets TCP en creation
  //Change le MSS qui est de 536 par defaut, le DelAckTimeout, et la version de TCP
  Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (MSS));//MSS: constante en debut du fichier - doit compte l'en-tete TCP 
  Config::SetDefault ("ns3::TcpSocket::DelAckTimeout",TimeValue(Seconds (0.25))); 
  Config::SetDefault ("ns3::TcpL4Protocol::SocketType", StringValue ("ns3::TcpHighSpeed"));
  
  /************************************ Application *************************************/


  //Recepteur: OBU
  videoPlayer thePlayer1(obu.Get (0), videoName);
  videoPlayer thePlayer2(obu.Get (1), videoName);
  videoPlayer thePlayer3(obu.Get (2), videoName);
  videoPlayer thePlayer4(obu.Get (3), videoName);
  videoPlayer thePlayer5(obu.Get (4), videoName);



  //Emetteur: tout est gere par la classe videoServerApp
  //videoServerApp myVideoApp(videoServer.Get (0));//Un 2nd arg peut contenir le nom de la video et un 3eme l'adresse IP de l'Obu
  //videoServerApp myVideoApp(videoServer.Get (0),"10.2.0.13");//Un 2nd arg peut contenir le nom de la video
  videoServerApp myVideoApp1(videoServer.Get (0),videoName, "10.2.0.4");
  videoServerApp myVideoApp2(videoServer.Get (0),videoName, "10.2.0.5");//2eme OBU
  videoServerApp myVideoApp3(videoServer.Get (0),videoName, "10.2.0.6");//3eme OBU
  videoServerApp myVideoApp4(videoServer.Get (0),videoName, "10.2.0.7");//4eme OBU
  videoServerApp myVideoApp5(videoServer.Get (0),videoName, "10.2.0.8");//5eme OBU

  //Un bon example de simulatorSchedule sur des "class member functon": https://www.nsnam.org/docs/release/3.7/doxygen/classns3_1_1_simulator.html
  //Un schedule pour chaque vehicule/videoApp
  Simulator::Schedule (Seconds (2.0), &videoServerApp::connectToObu, &myVideoApp1);
  Simulator::Schedule (Seconds (2.1), &videoServerApp::connectToObu, &myVideoApp2);//Attention au demarrage des apps (peut bugger quand demarre tous en meme temps)...
  Simulator::Schedule (Seconds (2.2), &videoServerApp::connectToObu, &myVideoApp3);
  Simulator::Schedule (Seconds (2.3), &videoServerApp::connectToObu, &myVideoApp4);
  Simulator::Schedule (Seconds (2.4), &videoServerApp::connectToObu, &myVideoApp5);

  Simulator::Stop (Seconds (tempsFinSimu+1.0));

  Simulator::Run ();
  Simulator::Destroy ();


  return 0;
}




